{   Public include file for the PICPRG library.
*
*   This library provides a procedural interface to the PICster programmer.
*   This programmer is intended for programming Micropchip PIC microcontrollers.
}
const
  picprg_subsys_k = -48;               {subsystem ID for this library}

  picprg_fwmin_k = 2;                  {minimum firmware version requirement}
  picprg_maxlen_cmd_k = 9;             {max bytes sent with any command}
  picprg_maxlen_rsp_k = 8;             {max bytes received in any response}
  picprg_maxconfig_k = 16;             {max addresses that store config bits}
  picprg_maxother_k = 16;              {max number of other programmable addresses}
{
*   Status values unique to the PICPRG subsystem.
}
  picprg_stat_badnbits_k = 1;          {illegal number of bits passed to routine}
  picprg_stat_wrongrsp_k = 2;          {wrong response routine for this command}
  picprg_stat_close_k = 3;             {this use of library is being closed}
  picprg_stat_arg_bad_k = 4;           {unexpected argument}
  picprg_stat_idid_k = 5;              {ID command within ID block in env file}
  picprg_stat_names2_k = 6;            {NAMES previously given this ID block}
  picprg_stat_badvddname_k = 7;        {unrecognized name in VDD command}
  picprg_stat_badfam_k = 8;            {bad PIC family type name}
  picprg_stat_idmissing_k = 9;         {required info missing from ID block}
  picprg_stat_badcmd_k = 11;           {bad env file command name}
  picprg_stat_tkextra_k = 12;          {extra token on env file line}
  picprg_stat_nidblock_k = 13;         {not within an ID block}
  picprg_stat_ideof_k = 14;            {ID block open at end of env file set}
  picprg_stat_idnmatch_k = 16;         {no match found for target chip ID}
  picprg_stat_namenmatch_k = 17;       {no actual name matches expected name}
  picprg_stat_barg_atline_k = 18;      {bad argument on specific line of file}
  picprg_stat_noid_k = 19;             {unable to get valid device ID}
  picprg_stat_unkfam_k = 20;           {unknown PIC family ID encountered}
  picprg_stat_nerase_k = 21;           {no specific erase routine installed}
  picprg_stat_dupadr_k = 22;           {duplicate address in list}
  picprg_stat_ovflcfg_k = 23;          {TINFO config list overflow}
  picprg_stat_ovfloth_k = 24;          {TINFO other list overflow}
  picprg_stat_nconfig_k = 25;          {library not configured}
  picprg_stat_nwrite_k = 26;           {no specific WRITE routine installed}
  picprg_stat_nread_k = 27;            {no specific READ routine installed}
  picprg_stat_cfgmix_k = 28;           {attempt to write config and non-config bits}
  picprg_stat_fwvers1_k = 29;          {firmware version is incompatible, single}
  picprg_stat_fwvers2_k = 30;          {firmware version is incompatible, range}

type
  picprg_p_t = ^picprg_t;              {pointer to library use state}

  picprg_reset_k_t = (                 {IDs for the possible reset algorithms}
    picprg_reset_none_k = 0,           {no algorithm, dummy routine used}
    picprg_reset_62x_k = 1,            {Vpp on before Vdd, like 16F62x}
    picprg_reset_18f_k = 2);           {Vdd on before Vpp, like 18Fxxx}

  picprg_write_k_t = (                 {IDs for the possible write algorithms}
    picprg_write_none_k = 0,           {no algorithm, dummy routine used}
    picprg_write_16f_k = 1,            {generic 16F PIC}
    picprg_write_12f6_k = 2,           {for 12F629, 12F675}
    picprg_write_18f_k = 3);           {generic 18F PIC}

  picprg_read_k_t = (                  {IDs for the possible read algorithms}
    picprg_read_none_k = 0,            {no algorithm, dummy routine used}
    picprg_read_16f_k = 1,             {generic 16F PIC}
    picprg_read_18f_k = 2);            {generic 18F PIC}

  picprg_org_k_t = (                   {IDs for organizations creating firmware}
    picprg_org_min_k = 1,              {organization ID with min allowed ordinal}
    picprg_org_official_k = 1,         {organization in charge of official releases}
    picprg_org_max_k = 254);           {organization ID with max allowed ordinal}

  picprg_space_k_t = (                 {IDs for the different target address spaces}
    picprg_space_prog_k,               {program memory address space}
    picprg_space_data_k);              {data (EEPROM) memory address space}

  picprg_pininfo_k_t = (               {individual flags from PINS command}
    picprg_pininfo_le18_k);            {target chip has 18 or fewer pins}
  picprg_pininfo_t = set of picprg_pininfo_k_t;

  picprg_chipid_t = 0 .. 16#FFFF;      {hard coded target chip device ID word}
  picprg_adr_t = 0 .. 16#FFFFFF;       {target device address}
  picprg_dat_t = 0 .. 16#FFFF;         {one target data word}

  picprg_datar_t =                     {array of data values for consecutive adr}
    array[0 .. 16#FFFFFF] of picprg_dat_t; {max size declared, not used directly}
  picprg_datar_p_t = ^picprg_datar_t;

  picprg_picfam_k_t = (                {IDs for the different PIC families}
    picprg_picfam_unknown_k,           {PIC family type is unknown}
    picprg_picfam_12f6xx_k,            {PIC 12F629, 12F675}
    picprg_picfam_16f62x_k,            {PIC 16F62x}
    picprg_picfam_16f87x_k,            {PIC 16F87x}
    picprg_picfam_18fxx2_k);           {PIC 18Fxx2 and 18Fxx8}

  picprg_pandat8_t =                   {data for 18xxx 8 byte panel writes}
    array[0 .. 7] of int8u_t;

  picprg_8byte_t =                     {buffer of 8 data bytes}
    array[0 .. 7] of int8u_t;

  picprg_fw_t = record                 {firmware version info}
    org: picprg_org_k_t;               {ID of the creating organization}
    cvlo: int8u_t;                     {lowest official vers backward compatible to}
    cvhi: int8u_t;                     {highest official vers compatible with}
    vers: int8u_t;                     {1-254 version number withing organization}
    info: sys_int_conv32_t;            {32 bit data private to the organization}
    end;

  picprg_org_t = record                {definition of one software creating org}
    name: string_var80_t;              {arbitrary name string}
    webpage: string_var1024_t;         {web page URL without http://, lower case}
    end;

  picprg_orgs_t =                      {info about all possible organization IDs}
    array[picprg_org_min_k .. picprg_org_max_k] of picprg_org_t;

  picprg_vddvals_t = record            {target chip Vdd voltage levels}
    low: real;                         {low voltage for verify}
    norm: real;                        {voltage for normal operations}
    high: real;                        {high voltage for verify}
    end;

  picprg_idname_p_t = ^picprg_idname_t;
  picprg_idname_t = record             {info for one name within ID block}
    next_p: picprg_idname_p_t;         {points to next name in chain for this block}
    name: string_var32_t;              {specific chip name, upper case}
    vdd: picprg_vddvals_t;             {the Vdd levels for this specific chip}
    end;

  picprg_adrent_p_t = ^picprg_adrent_t;
  picprg_adrent_t = record             {one entry in list of address specs}
    next_p: picprg_adrent_p_t;         {pointer to next list entry}
    adr: picprg_adr_t;                 {the address}
    mask: picprg_dat_t;                {mask of valid bits at this address}
    end;

  picprg_idblock_p_t = ^picprg_idblock_t;
  picprg_idblock_t = record            {info from one ID block in environment file}
    next_p: picprg_idblock_p_t;        {pointer to next ID block in chain}
    mask: picprg_chipid_t;             {mask for relevant bits in device ID word}
    id: picprg_chipid_t;               {ID word with unmasked bits set to 0}
    vdd: picprg_vddvals_t;             {default Vdd when no name specified}
    name_p: picprg_idname_p_t;         {pnt to start of specific names chain}
    rev_mask: picprg_chipid_t;         {mask for revision field within device ID}
    rev_shft: sys_int_machine_t;       {bits to shift revision field right}
    pins: sys_int_machine_t;           {number of pins}
    nprog: picprg_adr_t;               {number of program memory locations}
    ndat: picprg_adr_t;                {number of data memory (EEPROM) locations}
    maskprg: picprg_dat_t;             {mask for program memory data}
    maskdat: picprg_dat_t;             {mask for data memory data}
    datmap: picprg_adr_t;              {address where data memory mapped in HEX file}
    tprog: real;                       {seconds to wait after each program cycle}
    config_p: picprg_adrent_p_t;       {pointer to list of config word addresses}
    other_p: picprg_adrent_p_t;        {pointer to list of other valid addresses}
    fam: picprg_picfam_k_t;            {PIC family ID}
    end;

  picprg_env_t = record                {all the info read from PICPRG.ENV file set}
    org: picprg_orgs_t;                {info about all possible organization IDs}
    idblock_p: picprg_idblock_p_t;     {pointer to first ID block}
    end;
  picprg_env_p_t = ^picprg_env_t;

  picprg_tinfo_t = record              {info about the target chip}
    name: string_var32_t;              {specific name of the chip, like 16LF876}
    id: picprg_chipid_t;               {complete device ID word}
    rev_mask: picprg_chipid_t;         {mask for revision field within device ID}
    rev_shft: sys_int_machine_t;       {bits to shift revision field right}
    rev: sys_int_machine_t;            {revision number from device ID word}
    vdd: picprg_vddvals_t;             {Vdd levels for this chip}
    pins: sys_int_machine_t;           {number of pins}
    nprog: picprg_adr_t;               {number of program memory locations}
    ndat: picprg_adr_t;                {number of data memory (EEPROM) locations}
    maskprg: picprg_dat_t;             {mask for program memory data}
    maskdat: picprg_dat_t;             {mask for data memory data}
    datmap: picprg_adr_t;              {address where data memory mapped in HEX file}
    tprog: real;                       {seconds to wait after each program cycle}
    nconfig: sys_int_machine_t;        {number of CONFIG addresses in list}
    config_p: picprg_adrent_p_t;       {pointer to list of config word addresses}
    nother: sys_int_machine_t;         {number of OTHER addresses in list}
    other_p: picprg_adrent_p_t;        {pointer to list of other valid addresses}
    fam: picprg_picfam_k_t;            {PIC family ID}
    end;

  picprg_buf_cmd_t = record            {info about command bytes to send}
    nbuf: sys_int_machine_t;           {number of bytes}
    buf:                               {the bytes}
      array[1 .. picprg_maxlen_cmd_k] of int8u_t;
    end;

  picprg_buf_rsp_t = record            {info about response to one command}
    nresp: sys_int_machine_t;          {number of bytes expected}
    nbuf: sys_int_machine_t;           {number of bytes received so far}
    ack: boolean;                      {ACK received for this command}
    buf:                               {the received bytes}
      array[1 .. picprg_maxlen_rsp_k] of int8u_t;
    end;

  picprg_cmd_t = record                {info for one remote system command/response}
    send: picprg_buf_cmd_t;            {output bytes buffer and info}
    recv: picprg_buf_rsp_t;            {input bytes buffer and info}
    done: sys_sys_event_id_t;          {event that is signalled when command done}
    end;
  picprg_cmd_p_t = ^picprg_cmd_t;

  picprg_erase_p_t = ^procedure (      {subroutine to erase the target chip}
    in      pr_p: picprg_p_t;          {pointer to library use state}
    out     stat: sys_err_t);          {completion status}
    val_param;

  picprg_write_p_t = ^procedure (      {subroutine to write array to target chip}
    in      pr_p: picprg_p_t;          {pointer to library use state}
    in      adr: picprg_adr_t;         {starting address to write to}
    in      n: picprg_adr_t;           {number of locations to write to}
    in      dat: univ picprg_datar_t;  {array of data to write}
    in      mask: picprg_dat_t;        {mask for valid bits in each data word}
    out     stat: sys_err_t);          {completion status}
    val_param;

  picprg_read_p_t = ^procedure (       {subroutine to read array from target chip}
    in      pr_p: picprg_p_t;          {pointer to library use state}
    in      adr: picprg_adr_t;         {starting address to read from}
    in      n: picprg_adr_t;           {number of locations to read from}
    in      mask: picprg_dat_t;        {mask for valid data bits}
    out     dat: univ picprg_datar_t;  {the returned data array, unused bits 0}
    out     stat: sys_err_t);          {completion status}
    val_param;

  picprg_t = record                    {state for one use of the PICPRG library}
    sio: sys_int_machine_t;            {system serial line number}
    mem_p: util_mem_context_p_t;       {pointer to mem context for this library use}
    conn: file_conn_t;                 {connection to the serial line}
    ready: sys_sys_event_id_t;         {signalled when ready to send next command}
    cmd_p: picprg_cmd_p_t;             {points to current command, if any}
    cmd_next_p: picprg_cmd_p_t;        {points to next command after current}
    lock_cmd: sys_sys_threadlock_t;    {interlock for pending commands pointers}
    erase_p: picprg_erase_p_t;         {pointer to ERASE subroutine}
    write_p: picprg_write_p_t;         {pointer to ARRAY WRITE subroutine}
    read_p: picprg_read_p_t;           {pointer to ARRAY READ subroutine}
    env: picprg_env_t;                 {info from environment file set}
    thid_in: sys_sys_thread_id_t;      {ID of input receiving thread}
    id: picprg_chipid_t;               {full ID word read from target chip, init 0}
    id_p: picprg_idblock_p_t;          {pnt to target ID info, initially NIL}
    name_p: picprg_idname_p_t;         {pnt specific named target, initially NIL}
    space: picprg_space_k_t;           {ID for current target address space}
    quit: boolean;                     {trying to close}
    end;

procedure picprg_close (               {end a use of this library}
  in out  pr: picprg_t;                {library use state, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_config (              {configure library for specific target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  in      name: univ string_var_arg_t; {expected name, blank gets default for ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_erase (               {erase all erasable non-volatile target mem}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_fw_check (            {check firmware compatibility}
  in out  pr: picprg_t;                {state for this use of the library}
  in      fw: picprg_fw_t;             {firmware information}
  out     stat: sys_err_t);            {completion status, error if FW incompatible}
  val_param; extern;

procedure picprg_fw_show1 (            {show firmware: version, org name}
  in out  pr: picprg_t;                {state for this use of the library}
  in      fw: picprg_fw_t;             {firmware information}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_id (                  {get the hard coded ID of the target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  out     id: picprg_chipid_t;         {returned chip ID in low bits, 0 = none}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_init (                {initialize library state to defaults}
  out     pr: picprg_t);               {returned ready to pass to OPEN}
  val_param; extern;

procedure picprg_open (                {open a new use of this library}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_read (                {read data from target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  in      adr: picprg_adr_t;           {starting address to read from}
  in      n: picprg_adr_t;             {number of locations to read from}
  in      mask: picprg_dat_t;          {mask for valid data bits}
  out     dat: univ picprg_datar_t;    {the returned data array, unused bits 0}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_reset (               {reset the target chip and associated state}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_tinfo (               {get detailed info about the target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  out     tinfo: picprg_tinfo_t;       {returned detailed target chip info}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_write (               {write array of data to the target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  in      adr: picprg_adr_t;           {starting address to write to}
  in      n: picprg_adr_t;             {number of locations to write}
  in      dat: univ picprg_datar_t;    {array of data to write}
  in      mask: picprg_dat_t;          {mask for valid bits in each data word}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;
{
*   Command routines.  These routines each start a new command to be sent to
*   the remote system and return.  The caller must supply a CMD structure,
*   which will be initialized as needed.  The CMD structure is returned
*   and is used to track the progress of the command.  It must eventually
*   be passed to the PICPRG_RSP_xxx routine of the same name as the command.
*   These are described below.
}
procedure picprg_cmd_nop (             {send NOP command, just sends ACK back}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_off (             {turn off power to target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_pins (            {get info about target chip pins}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_send (            {send serial bits to the target}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      n: sys_int_machine_t;        {1-16 number of serial bits to send}
  in      dat: sys_int_conv16_t;       {the data bits to send, LSB first}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_recv (            {read serial data bits from the target}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      n: sys_int_machine_t;        {1-16 number of bits to read}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_clkh (            {set the serial clock line high}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_clkl (            {set the serial clock line low}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_dath (            {set the serial data line high}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_datl (            {set the serial data line low}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_datr (            {read the data line as driven by the target}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_tdrive (          {test whether target is driving data line}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_wait (            {guaranteed wait before next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      t: real;                     {time to wait in seconds, clipped and rounded}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_fwinfo (          {get firmware version and other info}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vddvals (         {set target chip Vdd levels}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      vlo: real;                   {low Vdd level, volts}
  in      vnr: real;                   {normal Vdd level, volts}
  in      vhi: real;                   {high Vdd level, volts}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vddlow (          {set target Vdd to the low level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vddnorm (         {set target Vdd to the normal level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vddhigh (         {set target Vdd to the high level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vddoff (          {set target Vdd to off (0 volts)}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vppon (           {turn on target programming voltage}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_vppoff (          {turn off target programming voltage}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_idreset (         {select reset algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      id: picprg_reset_k_t;        {reset algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_idwrite (         {select write algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      id: picprg_write_k_t;        {write algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_idread (          {select read algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      id: picprg_read_k_t;         {read algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_reset (           {reset target chip, ready for programming}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_test1 (           {send debugging TEST1 command}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_adr (             {set address of next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      adr: picprg_adr_t;           {target address for next operation}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_read (            {read from target, increment address}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_write (           {write to target, increment address}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      dat: picprg_dat_t;           {the data to write}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_tprog (           {set the programming write cycle time}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      t: real;                     {wait time in seconds}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_spprog (          {select program memory space}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_spdata (          {select data memory space}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_incadr (          {increment adr for next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_adrinv (          {invalidate target address assumption}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_pan18 (           {specialized 8 byte panel write for 18xxx}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  in      dat: picprg_pandat8_t;       {the 8 data bytes to write}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_rbyte8 (          {read low bytes of next 8 target words}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmd_writing (         {indicate the target is being written to}
  in out  pr: picprg_t;                {state for this use of the library}
  out     cmd: picprg_cmd_t;           {info about the command in progress}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;
{
*   Response routines.  These routines wait for the response to a command
*   to be received and return the response data, if any.  The CMD structure
*   must have been created by a PICPRG_CMD_xxx routine of the same name.
*   All resources allocated to the CMD struction are released, and the
*   CMD structure can be used again by passing it to another PICPRG_CMD_xxx
*   routine.
*
*   All these routines block execution until the expected response is received
*   from the remote unit.  However, multiple CMD structures may be outstanding,
*   meaning other PICPRG_CMD_xxx calls can be made before the response
*   routines for previous commands are called.  The PICPRG_CMD_xxx and
*   PICPRG_RSP_xxx routines may also be called from multiple threads.
}
procedure picprg_rsp_nop (             {wait for NOP command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_off (             {wait for OFF command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_pins (            {wait for PINS command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     pinfo: picprg_pininfo_t;     {returned info about target chip pins}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_send (            {wait for SEND command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_recv (            {wait for RECV command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     dat: sys_int_conv32_t;       {returned bits shfited into LSB, high zero}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_clkh (            {wait for CLKH command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_clkl (            {wait for CLKL command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_dath (            {wait for DATH command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_datl (            {wait for DATL command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_datr (            {wait for DATR command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     high: boolean;               {TRUE if data line was high, FALSE for low}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_tdrive (          {wait for TDRIVE command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     drive: boolean;              {TRUE iff target chip is driving data line}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_wait (            {wait for WAIT command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_fwinfo (          {wait for FWINFO command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     fwinfo: picprg_fw_t;         {returned info about programmer firmware}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vddvals (         {wait for VDDVALS command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vddlow (          {wait for VDDLOW command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vddnorm (         {wait for VDDNORM command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vddhigh (         {wait for VDDHIGH command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vddoff (          {wait for VDDOFF command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vppon (           {wait for VPPON command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_vppoff (          {wait for VPPOFF command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_idreset (         {wait for IDRESET command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_idwrite (         {wait for IDWRITE command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_idread (          {wait for IDREAD command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_reset (           {wait for RESET command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_test1 (           {wait for TEST1 command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_adr (             {wait for ADR command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_read (            {wait for READ command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     dat: picprg_dat_t;           {data read from the target}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_write (           {wait for WRITE command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_tprog (           {wait for TPROG command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_spprog (          {wait for SPPROG command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_spdata (          {wait for SPDATA command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_incadr (          {wait for INCADR command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_adrinv (          {wait for ADRINV command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_pan18 (           {wait for PAN18 command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_rbyte8 (          {wait for RBYTE8 command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     buf: univ picprg_8byte_t;    {returned array of 8 bytes}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_rsp_writing (         {wait for WRITING command completion}
  in out  pr: picprg_t;                {state for this use of the library}
  in out  cmd: picprg_cmd_t;           {info about the command, returned invalid}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;
{
*   Blocking command routines.  These are wrappers that first call the
*   PICPRG_CMD_xxx routine for the command, then the PICPRG_RSP_xxx routine.
*   These routines only return after the entire response to the command
*   has been received, thereby not allowing any overlap between commands.
*   These routines are therefore not recommended for a large number of
*   repetitive operations, such as writing to or reading from large sections
*   of the target chip program memory.
}
procedure picprg_cmdw_nop (            {send NOP command, just sends ACK back}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_off (            {turn off power to target chip}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_pins (           {get info about target chip pins}
  in out  pr: picprg_t;                {state for this use of the library}
  out     pinfo: picprg_pininfo_t;     {returned info about target chip pins}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_send (           {send serial bits to the target}
  in out  pr: picprg_t;                {state for this use of the library}
  in      n: sys_int_machine_t;        {1-16 number of serial bits to send}
  in      dat: sys_int_conv16_t;       {the data bits to send, LSB first}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_recv (           {read serial data bits from the target}
  in out  pr: picprg_t;                {state for this use of the library}
  in      n: sys_int_machine_t;        {1-16 number of bits to read}
  out     dat: sys_int_conv32_t;       {returned bits shfited into LSB, high zero}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_clkh (           {set the serial clock line high}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_clkl (           {set the serial clock line low}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_dath (           {set the serial data line high}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_datl (           {set the serial data line low}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_datr (           {read the data line as driven by the target}
  in out  pr: picprg_t;                {state for this use of the library}
  out     high: boolean;               {TRUE if data line was high, FALSE for low}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_tdrive (         {test whether target is driving data line}
  in out  pr: picprg_t;                {state for this use of the library}
  out     drive: boolean;              {TRUE iff target chip is driving data line}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_wait (           {guaranteed wait before next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  in      t: real;                     {time to wait in seconds, clipped and rounded}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_fwinfo (         {get firmware version and other info}
  in out  pr: picprg_t;                {state for this use of the library}
  out     fwinfo: picprg_fw_t;         {returned info about programmer firmware}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vddvals (        {set target chip Vdd levels}
  in out  pr: picprg_t;                {state for this use of the library}
  in      vlo: real;                   {low Vdd level, volts}
  in      vnr: real;                   {normal Vdd level, volts}
  in      vhi: real;                   {high Vdd level, volts}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vddlow (         {set target Vdd to the low level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vddnorm (        {set target Vdd to the normal level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vddhigh (        {set target Vdd to the high level}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vddoff (         {set target Vdd to off (0 volts)}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vppon (          {turn on target programming voltage}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_vppoff (         {turn off target programming voltage}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_idreset (        {select reset algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  in      id: picprg_reset_k_t;        {reset algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_idwrite (        {select write algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  in      id: picprg_write_k_t;        {write algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_idread (         {select read algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  in      id: picprg_read_k_t;         {read algorithm ID}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_reset (          {reset target chip, ready for programming}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_test1 (          {send debugging TEST1 command}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_adr (            {set address of next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  in      adr: picprg_adr_t;           {target address for next operation}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_read (           {read from target, increment address}
  in out  pr: picprg_t;                {state for this use of the library}
  out     dat: picprg_dat_t;           {data read from the target}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_write (          {write to target, increment address}
  in out  pr: picprg_t;                {state for this use of the library}
  in      dat: picprg_dat_t;           {the data to write}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_tprog (          {set the programming write cycle time}
  in out  pr: picprg_t;                {state for this use of the library}
  in      t: real;                     {wait time in seconds}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_spprog (         {select program memory space}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_spdata (         {select data memory space}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_incadr (         {increment adr for next target operation}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_adrinv (         {invalidate target address assumption}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_pan18 (          {specialized 8 byte panel write for 18xxx}
  in out  pr: picprg_t;                {state for this use of the library}
  in      dat: picprg_pandat8_t;       {the 8 data bytes to write}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_rbyte8 (         {read low bytes of next 8 target words}
  in out  pr: picprg_t;                {state for this use of the library}
  out     buf: univ picprg_8byte_t;    {returned array of 8 bytes}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;

procedure picprg_cmdw_writing (        {indicate the target is being written to}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;
{
*   For debugging and testing.
}
procedure picprg_18_test (             {test an PIC18 algorithm}
  in out  pr: picprg_t;                {state for this use of the library}
  out     stat: sys_err_t);            {completion status}
  val_param; extern;
